//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//	This version has been modified by fccHandler, September 2007

#include "stdafx.h"

#include <windows.h>
#include <vfw.h>
#include <vd2/system/Error.h>
#include <vd2/Dita/resources.h>

#include "gui.h"
#include "AudioSource.h"
#include "AVIReadHandler.h"

//////////////////////////////////////////////////////////////////////////////

extern HWND g_hWnd;		// TODO: Remove in 1.5.0

//////////////////////////////////////////////////////////////////////////////

namespace {
	enum { kVDST_AudioSource = 6 };

	enum {
		kVDM_TruncatedMP3FormatFixed,
		kVDM_VBRAudioDetected
	};
}

//////////////////////////////////////////////////////////////////////////////

AudioSourceWAV::AudioSourceWAV(const wchar_t *szFile, LONG inputBufferSize)
{
	MMIOINFO mmi;
	VDStringA filenameA(VDTextWToA(szFile));

	memset(&mmi,0,sizeof mmi);
	mmi.cchBuffer	= inputBufferSize;
	hmmioFile		= mmioOpen((char *)filenameA.c_str(), &mmi, MMIO_READ | MMIO_ALLOCBUF);

	if (!hmmioFile) {
		const char *pError = "An unknown error has occurred";
		switch(mmi.wErrorRet) {
		case MMIOERR_FILENOTFOUND:		pError = "File not found"; break;
		case MMIOERR_OUTOFMEMORY:		pError = "Out of memory"; break;
		case MMIOERR_ACCESSDENIED:		pError = "Access is denied"; break;
		case MMIOERR_CANNOTOPEN:		// fall through
		case MMIOERR_INVALIDFILE:		pError = "The file could not be opened"; break;
		case MMIOERR_NETWORKERROR:		pError = "A network error occurred while opening the file"; break;
		case MMIOERR_PATHNOTFOUND:		pError = "The file path does not exist"; break;
		case MMIOERR_SHARINGVIOLATION:	pError = "The file is currently in use"; break;
		case MMIOERR_TOOMANYOPENFILES:	pError = "Too many files are already open"; break;
		}
		throw MyError("Cannot open \"%s\": %s.", filenameA.c_str(), pError);
	}
}

AudioSourceWAV::~AudioSourceWAV() {
	mmioClose(hmmioFile, 0);
}

#ifdef VIRTUALDUB_MPEG2

int AudioSourceWAV::ac3_syncinfo(void *src, long *srate, int *brate, int *chans) {

	static const int rate[19] = {32,40,48,56,64,80,96,112,
		128,160,192,224,256,320,384,448,512,576,640};

	static const char lfeon[8] = {16,16,4,4,4,1,4,1};
	static const char chan[8] = {2,1,2,3,3,4,4,5};

	const unsigned char *buf = (unsigned char *)src;

	if (*(short *)src == 0x770B && buf[4] < 0xC0) {
		int frmsizecod = buf[4] & 63;
		if (frmsizecod < 38) {
			int bsid = buf[5] >> 3;
			if (bsid < 12) {
				int acmod = buf[6] >> 5;
				int bitrate = rate[frmsizecod >> 1];

				if (bsid > 8) bsid -= 8; else bsid = 0;
				if (chans) *chans = chan[acmod] + !!(buf[6] & lfeon[acmod]);
				if (brate) *brate = bitrate >> bsid;

				switch (buf[4] & 0xC0) {
					case 0:
						if (srate) *srate = 48000 >> bsid;
						return (4 * bitrate);
					case 0x40:
						if (srate) *srate = 44100 >> bsid;
						return (2 * (320 * bitrate / 147 + (frmsizecod & 1)));
					case 0x80:
						if (srate) *srate = 32000 >> bsid;
						return (6 * bitrate);
				}
			}
		}
	}
	return 0;
}

int AudioSourceWAV::mp3_syncinfo(void *src, long *srate, int *brate, int *chans) {

	// The following table is:
	//	bitrates for MPEG-1 Layer I
	//	bitrates for MPEG-1 Layer II
	//	bitrates for MPEG-1 Layer III
	//	bitrates for MPEG-2 Layer III

	static const unsigned short brates[4][16] = {
	    { 0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, 0 },
		{ 0, 32, 48, 56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384, 0 },
		{ 0, 32, 40, 48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 0 },
		{ 0,  8, 16, 24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160, 0 }
	};

	static const long srates[3] = { 44100, 48000, 32000 };

	int ver, layer, kbps;
	long samplerate;
	unsigned long hdr = *(unsigned long *)src;

	int fl = 0;

// ........ ........ 1111.... 11111111   sync (all bits set)
// ........ ........ ....x... ........   version (1=MPEG-1, 0=MPEG-2)
// ........ ........ .....xx. ........   layer (00=reserved,11=I,10=II,01=III)
// ........ ........ .......x ........   protected by CRC (0=yes, 1=no)
// ........ xxxx.... ........ ........   bitrate_index (can't be 1111)
// ........ ....xx.. ........ ........   sampling_freq (can't be 11)
// ........ ......x. ........ ........   padding_bit (0=no, 1=yes)
// ........ .......x ........ ........   private_bit (not used)
// xx...... ........ ........ ........   (00=stereo,01=joint,10=dual,11=mono)
// ..xx.... ........ ........ ........   mode_ext (only if joint stereo)
// ....x... ........ ........ ........   copyright_bit (0=no, 1=yes)
// .....x.. ........ ........ ........   original_or_copy (0=original, 1=copy)
// ......xx ........ ........ ........   emphasis (00=none,01=reserved,10=50/15ms,11=CCIT J.17)

// Start by verifying the fields...
	if ((hdr & 0xF0FF) != 0xF0FF) goto Abort;

	ver = 2 - ((hdr >> 11) & 1);	// 1 or 2
	layer = 4 - ((hdr >> 9) & 3);	// 1, 2, 3, or 4

	if (layer == 4) goto Abort;
	if (layer < 3 && ver != 1) goto Abort;

	kbps = (hdr >> 20) & 15;
	if (kbps == 0 || kbps == 15) goto Abort;

	if (layer < 3) {
		kbps = brates[layer - 1][kbps];
	} else {
		kbps = brates[ver + 1][kbps];
	}

	samplerate = (hdr >> 18) & 3;
	if (samplerate == 3) goto Abort;
	samplerate = srates[samplerate] / ver;

// Now we have a verified MPEG audio frame
// The return value is its size in bytes

	if (layer > 1) {
		fl = ((144000 / ver) * kbps) / samplerate;
		if (hdr & 0x20000) fl++;
	} else {
		fl = ((12000L / ver) * kbps) / samplerate;
		if (hdr & 0x20000) fl++;
		fl *= 4;
	}

	if (srate != NULL) *srate = samplerate;
	if (brate != NULL) *brate = kbps;
	if (chans != NULL) *chans = ((hdr >> 30) == 3)? 1: 2;

Abort:
	return fl;
}

void AudioSourceWAV::Detect() {
	BYTE buf[12];
	LONG lSeek2;
	iAudioType = 0;

	if (mmioRead(hmmioFile, (HPSTR)buf, 12) == 12) {

		if (*(DWORD *)buf == 'FFIR' && *(DWORD *)(buf+8) == 'EVAW') {
			iAudioType = 1;

		} else if (ac3_syncinfo(buf, NULL, NULL, NULL)) {
			chunkDATA.dwDataOffset = 0;
			iAudioType = 2;

		} else {
			LONG size = 0;

			if (buf[0] == 'I' && buf[1] == 'D' && buf[2] == '3') {
				// Damn ID3v2
				size = (buf[6]<<21)+(buf[7]<<14)+(buf[8]<<7)+buf[9];

				size += 10;
				if (mmioSeek(hmmioFile, size, SEEK_SET) != size) goto Abort;
				if (mmioRead(hmmioFile, (HPSTR)buf, 4) != 4) goto Abort;
			}

			if (mp3_syncinfo(buf, NULL, NULL, NULL)) {
				chunkDATA.dwDataOffset = size;
				iAudioType = 3;
			}
		}
	}

	lFileSize = mmioSeek(hmmioFile, 0, SEEK_END);
	lSeek2 = mmioSeek(hmmioFile, 0, SEEK_SET);

	if (lFileSize == -1 || lSeek2 == -1) {
Abort:
		lFileSize = 0;
		iAudioType = 0;
	}
}

#endif	// VIRTUALDUB_MPEG2

bool AudioSourceWAV::init() {

#ifdef VIRTUALDUB_MPEG2
	Detect();

	if (iAudioType == 1) {
#endif

	chunkRIFF.fccType = mmioFOURCC('W','A','V','E');
	if (MMSYSERR_NOERROR != mmioDescend(hmmioFile, &chunkRIFF, NULL, MMIO_FINDRIFF))
		return FALSE;

	chunkDATA.ckid = mmioFOURCC('f','m','t',' ');
	if (MMSYSERR_NOERROR != mmioDescend(hmmioFile, &chunkDATA, &chunkRIFF, MMIO_FINDCHUNK))
		return FALSE;

	if (!allocFormat(chunkDATA.cksize)) return FALSE;
	if (chunkDATA.cksize != mmioRead(hmmioFile, (char *)getWaveFormat(), chunkDATA.cksize))
		return FALSE;

	if (MMSYSERR_NOERROR != mmioAscend(hmmioFile, &chunkDATA, 0))
		return FALSE;

	chunkDATA.ckid = mmioFOURCC('d','a','t','a');
	if (MMSYSERR_NOERROR != mmioDescend(hmmioFile, &chunkDATA, &chunkRIFF, MMIO_FINDCHUNK))
		return FALSE;

	bytesPerSample	= getWaveFormat()->nBlockAlign; //getWaveFormat()->nAvgBytesPerSec / getWaveFormat()->nSamplesPerSec;
	mSampleFirst	= 0;
	mSampleLast		= chunkDATA.cksize / bytesPerSample;
	lCurrentSample	= -1;

	streamInfo.fccType					= streamtypeAUDIO;
	streamInfo.fccHandler				= 0;
	streamInfo.dwFlags					= 0;
	streamInfo.wPriority				= 0;
	streamInfo.wLanguage				= 0;
	streamInfo.dwInitialFrames			= 0;
	streamInfo.dwScale					= bytesPerSample;
	streamInfo.dwRate					= getWaveFormat()->nAvgBytesPerSec;
	streamInfo.dwStart					= 0;
	streamInfo.dwLength					= chunkDATA.cksize / bytesPerSample;
	streamInfo.dwSuggestedBufferSize	= 0;
	streamInfo.dwQuality				= 0xffffffff;
	streamInfo.dwSampleSize				= bytesPerSample;

	return TRUE;

#ifdef VIRTUALDUB_MPEG2
	}
	
	if (iAudioType == 2) {
		// Dolby Digital audio
		unsigned char buf[8];
		int brate, chans;
		long srate;

		WAVEFORMATEX *wfex;

		if (8 != mmioRead(hmmioFile, (HPSTR)buf, 8)) return FALSE;

		bytesPerSample = ac3_syncinfo(buf, &srate, &brate, &chans);
		if (!bytesPerSample) return FALSE;

		wfex = (WAVEFORMATEX *)allocFormat(sizeof(WAVEFORMATEX) * 2);
		if (!wfex) return FALSE;

		// Construct WAVEFORMATEX for AC-3
		wfex->wFormatTag		= 0x2000;
		wfex->nChannels			= (WORD)chans;
		wfex->nSamplesPerSec	= srate;
		wfex->nBlockAlign		= (WORD)bytesPerSample;
		wfex->nAvgBytesPerSec	= brate * 125UL;
		wfex->wBitsPerSample	= 0;

		// I think someone goofed this
		wfex->cbSize = sizeof(WAVEFORMATEX);
		memset((char *)wfex + sizeof(WAVEFORMATEX), 0, sizeof(WAVEFORMATEX));

		mSampleFirst = 0;
		mSampleLast = lFileSize / bytesPerSample;
		lCurrentSample = -1;

		streamInfo.fccType					= streamtypeAUDIO;
		streamInfo.fccHandler				= 0;
		streamInfo.dwCaps					= 0;
		streamInfo.wPriority				= 0;
		streamInfo.wLanguage				= 0;
		streamInfo.dwFlags					= 0;
		streamInfo.dwScale					= wfex->nBlockAlign;
		streamInfo.dwRate					= wfex->nAvgBytesPerSec;
		streamInfo.dwStart					= 0;
		streamInfo.dwLength					= (long)mSampleLast;
		streamInfo.dwInitialFrames			= 0;
		streamInfo.dwSuggestedBufferSize	= 0;
		streamInfo.dwQuality				= 0xffffffffL;
		streamInfo.dwSampleSize				= wfex->nBlockAlign;
		streamInfo.rcFrame.left				= 0;
		streamInfo.rcFrame.top				= 0;
		streamInfo.rcFrame.right			= 0;
		streamInfo.rcFrame.bottom			= 0;

		return TRUE;
	}

	if (iAudioType == 3) {
		// MPEG audio
		unsigned long hdr;
		int layer, brate, chans;
		long srate;

		if (mmioSeek(hmmioFile, chunkDATA.dwDataOffset, SEEK_SET) != chunkDATA.dwDataOffset) return FALSE;
		if (mmioRead(hmmioFile, (HPSTR)&hdr, 4) != 4) return FALSE;

		layer = 4 - ((hdr >> 9) & 3);

		if (layer < 3) {
			// MPEG audio, layer I or II...
			MPEG1WAVEFORMAT *wfex2;

			bytesPerSample = mp3_syncinfo(&hdr, &srate, &brate, &chans);
			if (!bytesPerSample) return FALSE;

			wfex2 = (MPEG1WAVEFORMAT *)allocFormat(sizeof(MPEG1WAVEFORMAT));
			if (!wfex2) return FALSE;

			wfex2->wfx.wFormatTag		= WAVE_FORMAT_MPEG;
			wfex2->wfx.nChannels		= chans;
			wfex2->wfx.nSamplesPerSec	= srate;
			wfex2->wfx.nAvgBytesPerSec	= brate * 125;
			wfex2->wfx.nBlockAlign		= (WORD)bytesPerSample;
			wfex2->wfx.wBitsPerSample	= 0;
			wfex2->wfx.cbSize			= 22;

			if (layer == 1) {
				wfex2->fwHeadLayer		= ACM_MPEG_LAYER1;
			} else {
				wfex2->fwHeadLayer		= ACM_MPEG_LAYER2;
			}

			wfex2->dwHeadBitrate		= brate * 1000;

// ........ ........ 1111.... 11111111   sync (all bits set)
// ........ ........ ....x... ........   version (1=MPEG-1, 0=MPEG-2)
// ........ ........ .....xx. ........   layer (00=reserved,11=I,10=II,01=III)
// ........ ........ .......x ........   protected by CRC (0=yes, 1=no)
// ........ xxxx.... ........ ........   bitrate_index (can't be 1111)
// ........ ....xx.. ........ ........   sampling_freq (can't be 11)
// ........ ......x. ........ ........   padding_bit (0=no, 1=yes)
// ........ .......x ........ ........   private_bit (not used)
// xx...... ........ ........ ........   (00=stereo,01=joint,10=dual,11=mono)
// ..xx.... ........ ........ ........   mode_ext (only if joint stereo)
// ....x... ........ ........ ........   copyright_bit (0=no, 1=yes)
// .....x.. ........ ........ ........   original_or_copy (0=original, 1=copy)
// ......xx ........ ........ ........   emphasis (00=none,01=reserved,10=50/15ms,11=CCIT J.17)

			switch ((hdr >> 30) & 3) {
				case 0: wfex2->fwHeadMode = ACM_MPEG_STEREO; break;
				case 1: wfex2->fwHeadMode = ACM_MPEG_DUALCHANNEL; break;
				case 2: wfex2->fwHeadMode = ACM_MPEG_JOINTSTEREO; break;
				case 3: wfex2->fwHeadMode = ACM_MPEG_SINGLECHANNEL; break;
			}

			wfex2->fwHeadModeExt		= (hdr >> 28) & 3;
			wfex2->wHeadEmphasis		= (hdr >> 24) & 3;

			wfex2->fwHeadFlags			= ACM_MPEG_ID_MPEG1;
			if (hdr & 0x00010000) wfex2->fwHeadFlags |= ACM_MPEG_PRIVATEBIT;
			if (hdr & 0x08000000) wfex2->fwHeadFlags |= ACM_MPEG_COPYRIGHT;
			if (hdr & 0x04000000) wfex2->fwHeadFlags |= ACM_MPEG_ORIGINALHOME;
			if (hdr & 0x00000100) wfex2->fwHeadFlags |= ACM_MPEG_PROTECTIONBIT;

			wfex2->dwPTSLow = 0;
			wfex2->dwPTSHigh = 0;

			mSampleFirst = 0;
			mSampleLast = lFileSize / bytesPerSample;
			lCurrentSample = -1;

			streamInfo.fccType				= streamtypeAUDIO;
			streamInfo.fccHandler			= 0;
			streamInfo.dwCaps				= 0;
			streamInfo.wPriority			= 0;
			streamInfo.wLanguage			= 0;
			streamInfo.dwFlags				= 0;
			streamInfo.dwScale				= wfex2->wfx.nBlockAlign;
			streamInfo.dwRate				= wfex2->wfx.nAvgBytesPerSec;
			streamInfo.dwStart				= 0;
			streamInfo.dwLength				= mSampleLast;
			streamInfo.dwInitialFrames		= 0;
			streamInfo.dwSuggestedBufferSize= 0;
			streamInfo.dwQuality			= 0xffffffffL;
			streamInfo.dwSampleSize			= wfex2->wfx.nBlockAlign;
			streamInfo.rcFrame.left			= 0;
			streamInfo.rcFrame.top			= 0;
			streamInfo.rcFrame.right		= 0;
			streamInfo.rcFrame.bottom		= 0;

			return TRUE;
		} 
		
		if (layer == 3) {
			MPEGLAYER3WAVEFORMAT *wfex3;

			bytesPerSample = mp3_syncinfo(&hdr, &srate, &brate, &chans);
			if (!bytesPerSample) return FALSE;

			wfex3 = (MPEGLAYER3WAVEFORMAT *)allocFormat(sizeof(MPEGLAYER3WAVEFORMAT));
			if (!wfex3) return FALSE;

			wfex3->wfx.wFormatTag		= WAVE_FORMAT_MPEGLAYER3;
			wfex3->wfx.nChannels		= chans;
			wfex3->wfx.nSamplesPerSec	= srate;
			wfex3->wfx.nBlockAlign		= (WORD)bytesPerSample;
			wfex3->wfx.nAvgBytesPerSec	= brate * 125UL;
			wfex3->wfx.wBitsPerSample	= 0;
			wfex3->wfx.cbSize			= MPEGLAYER3_WFX_EXTRA_BYTES;
			wfex3->wID					= MPEGLAYER3_ID_MPEG;
			wfex3->fdwFlags				= MPEGLAYER3_FLAG_PADDING_ISO;
			wfex3->nBlockSize			= bytesPerSample;
			wfex3->nFramesPerBlock		= 1;
			wfex3->nCodecDelay			= 0;	//1393?;

			mSampleFirst = 0;
			mSampleLast = lFileSize / bytesPerSample;
			lCurrentSample = -1;

			streamInfo.fccType				= streamtypeAUDIO;
			streamInfo.fccHandler			= 0;
			streamInfo.dwCaps				= 0;
			streamInfo.wPriority			= 0;
			streamInfo.wLanguage			= 0;
			streamInfo.dwFlags				= 0;
			streamInfo.dwScale				= wfex3->wfx.nBlockAlign;
			streamInfo.dwRate				= wfex3->wfx.nAvgBytesPerSec;
			streamInfo.dwStart				= 0;
			streamInfo.dwLength				= mSampleLast;
			streamInfo.dwInitialFrames		= 0;
			streamInfo.dwSuggestedBufferSize= 0;
			streamInfo.dwQuality			= 0xFFFFFFFF;
			streamInfo.dwSampleSize			= wfex3->wfx.nBlockAlign;
			streamInfo.rcFrame.left			= 0;
			streamInfo.rcFrame.top			= 0;
			streamInfo.rcFrame.right		= 0;
			streamInfo.rcFrame.bottom		= 0;

			return TRUE;
		}
	}

	return FALSE;
#endif	// VIRTUALDUB_MPEG2
}

int AudioSourceWAV::_read(VDPosition lStart, uint32 lCount, void *buffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead) {
	LONG lBytes = lCount * bytesPerSample;

	if (lBytes > cbBuffer) {
		lBytes = cbBuffer - cbBuffer % bytesPerSample;
		lCount = lBytes / bytesPerSample;
	}
	
	if (buffer) {
		if (lStart != lCurrentSample)
			if (-1 == mmioSeek(hmmioFile, chunkDATA.dwDataOffset + bytesPerSample*lStart, SEEK_SET))
				return AVIERR_FILEREAD;

		if (lBytes != mmioRead(hmmioFile, (char *)buffer, lBytes))
			return AVIERR_FILEREAD;

		lCurrentSample = lStart + lCount;
	}

	*lSamplesRead = lCount;
	*lBytesRead = lBytes;

	return AVIERR_OK;
}

///////////////////////////

AudioSourceAVI::AudioSourceAVI(IAVIReadHandler *pAVI, bool bAutomated) {
	pAVIFile	= pAVI;
	pAVIStream	= NULL;
	bQuiet = bAutomated;	// ugh, this needs to go... V1.5.0.
}

AudioSourceAVI::~AudioSourceAVI() {
	if (pAVIStream)
		delete pAVIStream;
}

bool AudioSourceAVI::init() {
	LONG format_len;

	pAVIStream = pAVIFile->GetStream(streamtypeAUDIO, 0);
	if (!pAVIStream) return FALSE;

	if (pAVIStream->Info(&streamInfo, sizeof streamInfo))
		return FALSE;

	pAVIStream->FormatSize(0, &format_len);

	if (!allocFormat(format_len)) return FALSE;

	if (pAVIStream->ReadFormat(0, getFormat(), &format_len))
		return FALSE;

	mSampleFirst = pAVIStream->Start();
	mSampleLast = pAVIStream->End();

	// Check for illegal (truncated) MP3 format.
	const WAVEFORMATEX *pwfex = getWaveFormat();

	if (pwfex->wFormatTag == WAVE_FORMAT_MPEGLAYER3 && format_len < sizeof(MPEGLAYER3WAVEFORMAT)) {
		MPEGLAYER3WAVEFORMAT wf;

		wf.wfx				= *pwfex;
		wf.wfx.cbSize		= MPEGLAYER3_WFX_EXTRA_BYTES;

		wf.wID				= MPEGLAYER3_ID_MPEG;

		// Attempt to detect the padding mode and block size for the stream.

		double byterate = wf.wfx.nAvgBytesPerSec;
		double fAverageFrameSize = 1152.0 * byterate / wf.wfx.nSamplesPerSec;

		int estimated_bitrate = (int)floor(0.5 + byterate * (1.0/1000.0)) * 8;
		double fEstimatedFrameSizeISO = 144000.0 * estimated_bitrate / wf.wfx.nSamplesPerSec;

		if (wf.wfx.nSamplesPerSec < 32000) {	// MPEG-2?
			fAverageFrameSize *= 0.5;
			fEstimatedFrameSizeISO *= 0.5;
		}

		double fEstimatedFrameSizePaddingOff = floor(fEstimatedFrameSizeISO);
		double fEstimatedFrameSizePaddingOn = fEstimatedFrameSizePaddingOff + 1.0;
		double fErrorISO = fabs(fEstimatedFrameSizeISO - fAverageFrameSize);
		double fErrorPaddingOn = fabs(fEstimatedFrameSizePaddingOn - fAverageFrameSize);
		double fErrorPaddingOff = fabs(fEstimatedFrameSizePaddingOff - fAverageFrameSize);

		if (fErrorISO <= fErrorPaddingOn)				// iso < on
			if (fErrorISO <= fErrorPaddingOff)			// iso < on, off
				wf.fdwFlags			= MPEGLAYER3_FLAG_PADDING_ISO;
			else										// off < iso < on
				wf.fdwFlags			= MPEGLAYER3_FLAG_PADDING_OFF;
		else											// on < iso
			if (fErrorPaddingOn <= fErrorPaddingOff)	// on < iso, off
				wf.fdwFlags			= MPEGLAYER3_FLAG_PADDING_ON;
			else										// off < on < iso
				wf.fdwFlags			= MPEGLAYER3_FLAG_PADDING_OFF;

		wf.nBlockSize		= (uint16)floor(0.5 + fAverageFrameSize);	// This is just a guess.  The MP3 codec always turns padding off, so I don't know whether this should be rounded up or not.
		wf.nFramesPerBlock	= 1;
		wf.nCodecDelay		= 1393;									// This is the number of samples padded by the compressor.  1393 is the value typically written by the codec.

		if (!allocFormat(sizeof wf))
			return FALSE;

		memcpy(getFormat(), &wf, sizeof wf);

		const int bad_len = format_len;
		const int good_len = sizeof(MPEGLAYER3WAVEFORMAT);
		VDLogAppMessage(kVDLogWarning, kVDST_AudioSource, kVDM_TruncatedMP3FormatFixed, 2, &bad_len, &good_len);
	}

	if (!bQuiet) {
		double mean, stddev, maxdev;

		if (pAVIStream->getVBRInfo(mean, stddev, maxdev)) {
			double meanOut = mean*0.001;
			double stddevOut = stddev*0.001;
			double maxdevOut = maxdev*1000.0;

			VDLogAppMessage(kVDLogWarning, kVDST_AudioSource, kVDM_VBRAudioDetected, 3, &maxdevOut, &meanOut, &stddevOut);
		}
	}

	return TRUE;
}

void AudioSourceAVI::Reinit() {
	pAVIStream->Info(&streamInfo, sizeof streamInfo);
	mSampleFirst = pAVIStream->Start();
	mSampleLast = pAVIStream->End();
}

int AudioSourceAVI::GetPreloadSamples() {
	long bytes, samples;
	if (pAVIStream->Read(mSampleFirst, AVISTREAMREAD_CONVENIENT, NULL, 0, &bytes, &samples))
		return 0;

	sint64 astart = pAVIStream->getSampleBytePosition(mSampleFirst);

	if (astart < 0)
		return (int)samples;

	IAVIReadStream *pVS = pAVIFile->GetStream(streamtypeVIDEO, 0);
	if (!pVS)
		return (int)samples;

	sint64 vstart = pVS->getSampleBytePosition(pVS->Start());
	delete pVS;

	if (vstart >= 0 && vstart < astart)
		return 0;

	return (int)samples;
}

bool AudioSourceAVI::isStreaming() {
	return pAVIStream->isStreaming();
}

void AudioSourceAVI::streamBegin(bool fRealTime, bool bForceReset) {
	pAVIStream->BeginStreaming(mSampleFirst, mSampleLast, fRealTime ? 1000 : 2000);
}

void AudioSourceAVI::streamEnd() {
	pAVIStream->EndStreaming();

}

bool AudioSourceAVI::_isKey(VDPosition lSample) {
	return pAVIStream->IsKeyFrame(lSample);
}

int AudioSourceAVI::_read(VDPosition lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lpBytesRead, uint32 *lpSamplesRead) {
	int err;
	long lBytes, lSamples;

	// There are some video clips roaming around with truncated audio streams
	// (audio streams that state their length as being longer than they
	// really are).  We use a kludge here to get around the problem.

	err = pAVIStream->Read(lStart, lCount, lpBuffer, cbBuffer, &lBytes, &lSamples);

	if (lpBytesRead)
		*lpBytesRead = lBytes;
	if (lpSamplesRead)
		*lpSamplesRead = lSamples;

	if (err != AVIERR_FILEREAD)
		return err;

	// Suspect a truncated stream.
	//
	// AVISTREAMREAD_CONVENIENT will tell us if we're actually encountering a
	// true read error or not.  At least for the AVI handler, it returns
	// AVIERR_ERROR if we've broached the end.  

	*lpBytesRead = *lpSamplesRead = 0;

	while(lCount > 0) {
		err = pAVIStream->Read(lStart, AVISTREAMREAD_CONVENIENT, NULL, 0, &lBytes, &lSamples);

		if (err)
			return 0;

		if (!lSamples) return AVIERR_OK;

		if (lSamples > lCount) lSamples = lCount;

		err = pAVIStream->Read(lStart, lSamples, lpBuffer, cbBuffer, &lBytes, &lSamples);

		if (err)
			return err;

		lpBuffer = (LPVOID)((char *)lpBuffer + lBytes);
		cbBuffer -= lBytes;
		lCount -= lSamples;

		*lpBytesRead += lBytes;
		*lpSamplesRead += lSamples;
	}

	return AVIERR_OK;
}

VDPosition AudioSourceAVI::TimeToPositionVBR(VDTime us) const {
	return pAVIStream->TimeToPosition(us);
}

VDTime AudioSourceAVI::PositionToTimeVBR(VDPosition samples) const {
	return pAVIStream->PositionToTime(samples);
}

///////////////////////////////////////////////////////////////////////////

AudioSourceDV::AudioSourceDV(IAVIReadStream *pStream, bool bAutomated)
	: mpStream(pStream)
	, mLastFrame(-1)
{
	bQuiet = bAutomated;	// ugh, this needs to go... V1.5.0.
}

AudioSourceDV::~AudioSourceDV() {
	if (mpStream)
		delete mpStream;
}

bool AudioSourceDV::init() {
	LONG format_len;

	mpStream->FormatSize(0, &format_len);

	if (!allocFormat(sizeof(WAVEFORMATEX)))
		return false;

	WAVEFORMATEX *pwfex = (WAVEFORMATEX *)getFormat();

	// fetch AAUX packet from stream format and determine sampling rate
	long size;
	if (FAILED(mpStream->FormatSize(0, &size)))
		return false;

	if (size < 24)
		return false;

	vdblock<uint8> format(size);
	if (FAILED(mpStream->ReadFormat(0, format.data(), &size)))
		return false;

	uint8 aaux_as_pc4 = format[3];
	uint8 vaux_vs_pc3 = format[18];

	// Sometimes the values in the DVINFO block are wrong, so attempt
	// to extract from the first frame instead.
	const VDPosition streamStart = mpStream->Start();
	const VDPosition streamEnd = mpStream->End();
	
	if (streamEnd > streamStart) {
		long bytes, samples;

		if (!mpStream->Read(streamStart, 1, NULL, 0, &bytes, &samples) && bytes >= 120000) {
			vdblock<uint8> tmp(bytes);

			if (!mpStream->Read(streamStart, 1, tmp.data(), tmp.size(), &bytes, &samples)) {
				aaux_as_pc4 = tmp[80*(3*150 + 6) + 7];		// DIF sequence 3, block 6, AAUX pack 0
				vaux_vs_pc3 = tmp[80*(1*150 + 3) + 6];		// DIF sequence 1, block 3, VAUX pack 0
			}
		}
	}

	bool isPAL = 0 != (vaux_vs_pc3 & 0x20);

	sint32 samplingRate;

	switch(aaux_as_pc4 & 0x38) {
	case 0x00:
		samplingRate = 48000;
		mSamplesPerSet		= isPAL ? 19200 : 16016;
		break;
	case 0x08:
		samplingRate = 44100;
		mSamplesPerSet		= isPAL ? 17640 : 14715;
		break;
	case 0x10:
		samplingRate = 32000;
		mSamplesPerSet		= isPAL ? 12800 : 10677;
		break;
	default:
		return false;
	}

	// check for 12-bit quantization
	unsigned bytesPerSample = (aaux_as_pc4 & 7) ? 3 : 2;

	mTempBuffer.resize(isPAL ? 144000 : 120000);

	pwfex->wFormatTag		= WAVE_FORMAT_PCM;
	pwfex->nChannels		= 2;
	pwfex->nSamplesPerSec	= samplingRate;
	pwfex->nAvgBytesPerSec	= samplingRate*4;
	pwfex->nBlockAlign		= 4;
	pwfex->wBitsPerSample	= 16;
	pwfex->cbSize			= 0;

	mGatherTab.resize(1960);
	memset(mGatherTab.data(), 0, mGatherTab.size() * sizeof mGatherTab[0]);

	if (isPAL) {
		for(int i=0; i<1944; ++i) {
			int dif_sequence	= ((i/3)+2*(i%3))%6;
			int dif_block		= 6 + 16*(3*(i%3) + ((i%54)/18));
			int byte_offset		= 8 + bytesPerSample*(i/54);

			mGatherTab[i] = 12000*dif_sequence + 80*dif_block + byte_offset;
		}

		mRightChannelOffset = 12000*6;	// left channel is first 6 DIF sequences
	} else {
		for(int i=0; i<1620; ++i) {
			int dif_sequence	= ((i/3)+2*(i%3))%5;
			int dif_block		= 6 + 16*(3*(i%3) + ((i%45)/15));
			int byte_offset		= 8 + bytesPerSample*(i/45);

			mGatherTab[i] = 12000*dif_sequence + 80*dif_block + byte_offset;
		}

		mRightChannelOffset	= 12000*5;	// left channel is first 5 DIF sequences
	}

	mpStream->Info(&streamInfo, sizeof streamInfo);

	// wonk most of the stream values since they're not appropriate for audio
	streamInfo.fccType		= streamtypeAUDIO;
	streamInfo.fccHandler	= 0;
	streamInfo.dwStart		= VDRoundToInt((double)streamInfo.dwScale / streamInfo.dwRate * samplingRate);
	streamInfo.dwRate		= pwfex->nAvgBytesPerSec;
	streamInfo.dwScale		= pwfex->nBlockAlign;
	streamInfo.dwInitialFrames			= 0;
	streamInfo.dwSuggestedBufferSize	= 0;
	streamInfo.dwQuality				= (DWORD)-1;
	streamInfo.dwSampleSize				= pwfex->nBlockAlign;
	memset(&streamInfo.rcFrame, 0, sizeof streamInfo.rcFrame);

	Reinit();

	return true;
}

void AudioSourceDV::Reinit() {
	const VDPosition start = mpStream->Start();
	const VDPosition end = mpStream->End();

	mRawStart		= start;
	mRawEnd			= end;
	mRawFrames		= end - start;
	mSampleFirst	= (start * mSamplesPerSet) / 10;
	mSampleLast		= (start+end * mSamplesPerSet) / 10;

	streamInfo.dwLength	= mSampleLast - mSampleFirst;

	FlushCache();
}

bool AudioSourceDV::isStreaming() {
	return mpStream->isStreaming();
}

void AudioSourceDV::streamBegin(bool fRealTime, bool bForceReset) {
	mpStream->BeginStreaming(mSampleFirst, mSampleLast, fRealTime ? 1000 : 2000);
}

void AudioSourceDV::streamEnd() {
	mpStream->EndStreaming();

}

bool AudioSourceDV::_isKey(VDPosition lSample) {
	return true;
}

int AudioSourceDV::_read(VDPosition lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lpBytesRead, uint32 *lpSamplesRead) {
	if (lpBuffer && cbBuffer < 4)
		return AVIERR_BUFFERTOOSMALL;

	if (lCount == AVISTREAMREAD_CONVENIENT)
		lCount = mSamplesPerSet;

	VDPosition baseSet = lStart / mSamplesPerSet;
	uint32 offset = (uint32)(lStart % mSamplesPerSet);

	if (lCount > mSamplesPerSet - offset)
		lCount = mSamplesPerSet - offset;

	if (lpBuffer && lCount > (cbBuffer>>2))
		lCount = cbBuffer>>2;

	if (lpBuffer) {
		const CacheLine *pLine = LoadSet(baseSet);
		if (!pLine)
			return AVIERR_FILEREAD;

		memcpy(lpBuffer, (char *)pLine->mResampledData + offset*4, 4*lCount);
	}

	if (lpBytesRead)
		*lpBytesRead = lCount * 4;
	if (lpSamplesRead)
		*lpSamplesRead = lCount;

	return AVIERR_OK;
}

const AudioSourceDV::CacheLine *AudioSourceDV::LoadSet(VDPosition setpos) {
	// For now we will be lazy and just direct map the cache.
	unsigned line = (unsigned)setpos & 3;

	CacheLine& cline = mCache[line];

	if (mCacheLinePositions[line] != setpos) {
		mCacheLinePositions[line] = -1;

		// load up to 10 frames and linearize the raw data
		sint32 setSize = mSamplesPerSet;
		VDPosition pos = mRawStart + 10 * setpos;
		VDPosition limit = pos + 10;
		if (limit > mRawEnd) {
			limit = mRawEnd;

			setSize = mSamplesPerSet * ((sint32)limit - (sint32)pos) / 10;
		}
		cline.mRawSamples = 0;

		uint8 *dst = (uint8 *)cline.mRawData;

		while(pos < limit) {
			long bytes, samples;

			int err = mpStream->Read(pos++, 1, mTempBuffer.data(), mTempBuffer.size(), &bytes, &samples);
			if (err)
				return NULL;

			if (!bytes) {
				uint32 n = mSamplesPerSet / 10;

				if (cline.mRawSamples+n >= sizeof cline.mRawData / sizeof cline.mRawData[0]) {
					VDDEBUG("AudioSourceDV: Sample count overflow!\n");
					VDASSERT(false);
					break;
				}

				cline.mRawSamples += n;

				memset(dst, 0, n*4);
				dst += n*4;
			} else {
				if (bytes != mTempBuffer.size())
					return NULL;

				const uint8 *pAAUX = &mTempBuffer[80*(1*150 + 6) + 3];

				uint8 vaux_vs_pc3 = mTempBuffer[80*(1*150 + 3) + 6];		// DIF sequence 1, block 3, VAUX pack 0

				bool isPAL = 0 != (vaux_vs_pc3 & 0x20);

				uint32 minimumFrameSize;

				switch(pAAUX[4] & 0x38) {
				case 0x00:
					minimumFrameSize	= isPAL ? 1896 : 1580;
					break;
				case 0x08:
					minimumFrameSize	= isPAL ? 1742 : 1452;
					break;
				case 0x10:
					minimumFrameSize	= isPAL ? 1264 : 1053;
					break;
				default:
					return NULL;
				}

				const uint32 n = minimumFrameSize + (pAAUX[1] & 0x3f);

				if (cline.mRawSamples+n >= sizeof cline.mRawData / sizeof cline.mRawData[0]) {
					VDDEBUG("AudioSourceDV: Sample count overflow!\n");
					VDASSERT(false);
					break;
				}

				cline.mRawSamples += n;

				if ((pAAUX[4] & 7) == 1) {	// 1: 12-bit nonlinear
					const uint8 *src0 = (const uint8 *)&mTempBuffer[0];
					const ptrdiff_t rightOffset = mRightChannelOffset;
					sint32 *pOffsets = mGatherTab.data();

					uint16 *dst16 = (uint16 *)dst;
					dst += 4*n;

					for(int i=0; i<n; ++i) {
						const ptrdiff_t pos = *pOffsets++;
						const uint8 *srcF = src0 + pos;
						const uint8 *srcR = srcF + rightOffset;

						// Convert 12-bit nonlinear (one's complement floating-point) sample to 16-bit linear.
						// This value is a 1.3.8 floating-point value similar to IEEE style, except that the
						// sign is represented via 1's-c rather than sign-magnitude.
						//
						// Thus, 0000..00FF are positive denormals, 8000 is the largest negative value, etc.

						sint32 vL = ((sint32)srcF[0]<<4) + (srcF[2]>>4);		// reconstitute left 12-bit value
						sint32 vR = ((sint32)srcF[1]<<4) + (srcF[2] & 15);		// reconstitute right 12-bit value

						static const sint32 addend[16]={
							-0x0000 << 0,
							-0x0000 << 0,
							-0x0100 << 1,
							-0x0200 << 2,
							-0x0300 << 3,
							-0x0400 << 4,
							-0x0500 << 5,
							-0x0600 << 6,

							-0x09ff << 6,
							-0x0aff << 5,
							-0x0bff << 4,
							-0x0cff << 3,
							-0x0dff << 2,
							-0x0eff << 1,
							-0x0fff << 0,
							-0x0fff << 0,
						};
						static const int	shift [16]={0,0,1,2,3,4,5,6,6,5,4,3,2,1,0,0};

						int expL = vL >> 8;
						int expR = vR >> 8;

						dst16[0] = (vL << shift[expL]) + addend[expL];
						dst16[1] = (vR << shift[expR]) + addend[expR];
						dst16 += 2;
					}
				} else {					// 0: 16-bit linear
					const uint8 *src0 = (const uint8 *)&mTempBuffer[0];
					const ptrdiff_t rightOffset = mRightChannelOffset;
					sint32 *pOffsets = mGatherTab.data();

					for(int i=0; i<n; ++i) {
						const ptrdiff_t pos = *pOffsets++;
						const uint8 *srcL = src0 + pos;
						const uint8 *srcR = srcL + rightOffset;

						// convert big-endian sample to little-endian
						dst[0] = srcL[1];
						dst[1] = srcL[0];
						dst[2] = srcR[1];
						dst[3] = srcR[0];
						dst += 4;
					}
				}
			}
		}

		// resample if required
		if (cline.mRawSamples == setSize) {
			// no resampling required -- straight copy
			memcpy(cline.mResampledData, cline.mRawData, 4*setSize);
		} else {
			const sint16 *src = &cline.mRawData[0][0];
			      sint16 *dst = &cline.mResampledData[0][0];

			// copy first sample
			dst[0] = src[0];
			dst[1] = src[1];
			dst += 2;

			// linearly interpolate middle samples
			uint32 dudx = ((cline.mRawSamples-1) << 16) / (setSize-1);
			uint32 u = dudx + (dudx >> 1) - 0x8000;

			VDASSERT((sint32)u >= 0);

			unsigned n = setSize-2;
			do {
				const sint16 *src2 = src + (u>>16)*2;
				sint32 f = (u>>4)&0xfff;

				u += dudx;

				sint32 a0 = src2[0];
				sint32 b0 = src2[1];
				sint32 da = src2[2] - a0;
				sint32 db = src2[3] - b0;

				dst[0] = a0 + ((da*f + 0x800) >> 12);
				dst[1] = b0 + ((db*f + 0x800) >> 12);
				dst += 2;
			} while(--n);

			VDASSERT(((u - dudx) >> 16) < cline.mRawSamples);

			// copy last sample
			dst[0] = src[cline.mRawSamples*2-2];
			dst[1] = src[cline.mRawSamples*2-1];
		}

		VDDEBUG("AudioSourceDV: Loaded cache line %u for %u raw samples (%u samples expected)\n", (unsigned)setpos*10, cline.mRawSamples, setSize);

		mCacheLinePositions[line] = setpos;
	}

	return &cline;
}

void AudioSourceDV::FlushCache() {
	for(int i=0; i<kCacheLines; ++i)
		mCacheLinePositions[i] = -1;
}
